@@ -22,15 +22,14 @@
 #include "config.h"
 #include <string.h>
 #include <libxml/HTMLparser.h>
+#include <glib/gi18n.h>
 #include <webkit/webkit.h>
 #include <dbus/dbus-glib.h>
 #include <libsoup/soup-gnome-features.h>
 #include <libsoup/soup-uri.h>
-#include <gnome-keyring.h>
+#include <libsyncdaemon/libsyncdaemon.h>
 #include "oauth.h"
 #include "u1-music-store.h"
-#include "authentication.h"
-#include "syncdaemon.h"
 #include "u1-marshal.h"
 
 #define U1_MUSIC_LIBRARY_LOCATION  ".ubuntuone/Purchased from Ubuntu One"
@@ -48,8 +47,11 @@
 #define U1_CONNECTING_PAGE "connecting.html" 
 #define U1_INITIAL_PAGE "<html><body>Loading Ubuntu One music store</body></html>"
 
+#define MAXIMUM_AUTH_RETRIES 5
+
 struct _U1MusicStorePrivate {
 	DBusGConnection *bus;
+	SyncdaemonDaemon *syncdaemon;
 	gchar *base_url;
 
 	GtkWidget *scroll;
@@ -61,6 +63,8 @@ struct _U1MusicStorePrivate {
 	guint watch_id;
 	guint idle_cb;
 	GHashTable *watched_downloads;
+
+	gint auth_retries;
 };
 
 G_DEFINE_TYPE(U1MusicStore, u1_music_store, GTK_TYPE_VBOX)
@@ -80,36 +84,26 @@ u1_music_store_finalize (GObject *object)
 {
 	U1MusicStore *music_store = U1_MUSIC_STORE (object);
 
-	syncdaemon_finalize ();
-
 	if (music_store->priv != NULL) {
-		if (music_store->priv->idle_cb != 0) {
+		if (music_store->priv->idle_cb != 0)
 			g_source_remove (music_store->priv->idle_cb);
-			music_store->priv->idle_cb = 0;
-		}
 
-		if (music_store->priv->watch_id != 0) {
+		if (music_store->priv->watch_id != 0)
 			g_source_remove (music_store->priv->watch_id);
-			music_store->priv->watch_id = 0;
-		}
 
-		if (music_store->priv->watched_downloads != NULL) {
+		if (music_store->priv->watched_downloads != NULL)
 			g_hash_table_destroy (music_store->priv->watched_downloads);
-			music_store->priv->watched_downloads = NULL;
-		}
 
-		if (music_store->priv->base_url != NULL) {
+		if (music_store->priv->base_url != NULL)
 			g_free (music_store->priv->base_url);
-			music_store->priv->base_url;
-		}
 
-		if (music_store->priv->bus != NULL) {
+		if (music_store->priv->bus != NULL)
 			dbus_g_connection_unref (music_store->priv->bus);
-			music_store->priv->bus = NULL;
-		}
+
+		if (music_store->priv->syncdaemon != NULL)
+			g_object_unref (G_OBJECT (music_store->priv->syncdaemon));
 
 		g_free (music_store->priv);
-		music_store->priv = NULL;
 	}
 
 	G_OBJECT_CLASS (u1_music_store_parent_class)->finalize (object);
@@ -193,10 +187,16 @@ load_internal_html_page (WebKitWebView *web_view, const gchar *file_name, const
 }
 
 static void
-parse_oauth_string (const gchar *string, gchar **oauth_token, gchar **oauth_token_secret)
+parse_oauth_string (const gchar *string,
+		    gchar **oauth_consumer_token,
+		    gchar **oauth_consumer_secret,
+		    gchar **oauth_token,
+		    gchar **oauth_token_secret)
 {
 	gchar **items;
 
+	*oauth_consumer_token = NULL;
+	*oauth_consumer_secret = NULL;
 	*oauth_token = NULL;
 	*oauth_token_secret = NULL;
 
@@ -213,6 +213,10 @@ parse_oauth_string (const gchar *string, gchar **oauth_token, gchar **oauth_toke
 					*oauth_token = g_strdup (kv_pair[1]);
 				else if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_token_secret") == 0)
 					*oauth_token_secret = g_strdup (kv_pair[1]);
+				else if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_consumer_token") == 0)
+					*oauth_consumer_token = g_strdup (kv_pair[1]);
+				else if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_consumer_secret") == 0)
+					*oauth_consumer_secret = g_strdup (kv_pair[1]);
 
 				g_strfreev (kv_pair);
 			}
@@ -223,29 +227,23 @@ parse_oauth_string (const gchar *string, gchar **oauth_token, gchar **oauth_toke
 }
 
 static void
-get_credentials_from_keyring (gchar **oauth_token, gchar **oauth_token_secret)
+get_credentials (U1MusicStore *music_store,
+		 gchar **oauth_consumer_token,
+		 gchar **oauth_consumer_secret,
+		 gchar **oauth_token,
+		 gchar **oauth_token_secret)
 {
-	GnomeKeyringAttributeList *attrs;
-	GnomeKeyringResult result;
-	GList *items_found;
+	SyncdaemonCredentials *credentials;
 
-	*oauth_token = *oauth_token_secret = NULL;
+	*oauth_consumer_token = *oauth_consumer_secret = *oauth_token = *oauth_token_secret = NULL;
 
 	/* Get the OAuth token from the keyring */
-	attrs = gnome_keyring_attribute_list_new ();
-	gnome_keyring_attribute_list_append_string (attrs, "ubuntuone-realm", "https://ubuntuone.com");
-	gnome_keyring_attribute_list_append_string (attrs, "oauth-consumer-key", "ubuntuone");
-
-	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
-						attrs, &items_found);
-	if (result == GNOME_KEYRING_RESULT_OK && items_found != NULL) {
-		GnomeKeyringFound *first_item = (GnomeKeyringFound *) items_found->data;
-					
-		parse_oauth_string (first_item->secret, oauth_token, oauth_token_secret);
-		gnome_keyring_found_list_free (items_found);
-
-		/* Connect to syncdaemon now that we have tokens */
-		syncdaemon_connect ();
+	if ((credentials = syncdaemon_authentication_find_credentials (
+		     syncdaemon_daemon_get_authentication (music_store->priv->syncdaemon)))) {
+		*oauth_consumer_token = g_strdup (syncdaemon_credentials_get_consumer_key (credentials));
+		*oauth_consumer_secret = g_strdup (syncdaemon_credentials_get_consumer_secret (credentials));
+		*oauth_token = g_strdup (syncdaemon_credentials_get_token (credentials));
+		*oauth_token_secret = g_strdup (syncdaemon_credentials_get_token_secret (credentials));
 	}
 }
 
@@ -257,7 +255,7 @@ typedef struct {
 static gchar *
 get_url_to_use (U1MusicStore *music_store)
 {
-	gchar *oauth_token = NULL, *oauth_token_secret = NULL, *real_url;
+	gchar *oauth_consumer_token, *oauth_consumer_secret, *oauth_token = NULL, *oauth_token_secret = NULL, *real_url;
 	const gchar *url_to_use;
 
 	/* If U1MUSICSTOREURL is defined, use that instead of the real URL */
@@ -279,7 +277,11 @@ get_url_to_use (U1MusicStore *music_store)
 			gsize len;
 
 			if (g_file_get_contents (oauthfile, &oauthstring, &len, NULL)) {
-				parse_oauth_string (oauthstring, &oauth_token, &oauth_token_secret);
+				parse_oauth_string (oauthstring,
+						    &oauth_consumer_token,
+						    &oauth_consumer_secret,
+						    &oauth_token,
+						    &oauth_token_secret);
 				g_free (oauthstring);
 			}
 		} else {
@@ -289,8 +291,13 @@ get_url_to_use (U1MusicStore *music_store)
 						     "text/html", "utf-8", "file:///");
 			return NULL;
 		}
-	} else
-		get_credentials_from_keyring (&oauth_token, &oauth_token_secret);
+	} else {
+		get_credentials (music_store,
+				 &oauth_consumer_token,
+				 &oauth_consumer_secret,
+				 &oauth_token,
+				 &oauth_token_secret);
+	}
 
 	if (oauth_token == NULL || oauth_token_secret == NULL)
 		real_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_NOT_LOGGED_IN_STORE_URL);
@@ -298,11 +305,17 @@ get_url_to_use (U1MusicStore *music_store)
 		gchar *s = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_STORE_URL);
 
 		real_url = oauth_sign_url2 (s, NULL, OA_HMAC, "GET",
-					    "ubuntuone", "hammertime",
+					    oauth_consumer_token, oauth_consumer_secret,
 					    oauth_token, oauth_token_secret);
 		g_free (s);
+
+		/* Connect to syncdaemon now that we have tokens */
+		if (syncdaemon_daemon_is_ready (music_store->priv->syncdaemon))
+			syncdaemon_daemon_connect (music_store->priv->syncdaemon);
 	}
 
+	g_free (oauth_consumer_token);
+	g_free (oauth_consumer_secret);
 	g_free (oauth_token);
 	g_free (oauth_token_secret);
 
@@ -312,13 +325,18 @@ get_url_to_use (U1MusicStore *music_store)
 
 
 static void
-got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar *consumer_key, gpointer user_data)
+got_new_credentials_cb (SyncdaemonAuthentication *auth, SyncdaemonCredentials *credentials, gpointer user_data)
 {
-	gchar *oauth_token, *oauth_token_secret, *url_to_use, *real_url;
+	gchar *oauth_consumer_token, *oauth_consumer_secret, *oauth_token, *oauth_token_secret, *url_to_use, *real_url;
 	gchar **uri_parts;
 	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;
 
-	get_credentials_from_keyring (&oauth_token, &oauth_token_secret);
+	g_debug ("Got new credentials from ubuntu-sso");
+
+	oauth_consumer_token = g_strdup (syncdaemon_credentials_get_consumer_key (credentials));
+	oauth_consumer_secret = g_strdup (syncdaemon_credentials_get_consumer_secret (credentials));
+	oauth_token = g_strdup (syncdaemon_credentials_get_token (credentials));
+	oauth_token_secret = g_strdup (syncdaemon_credentials_get_token_secret (credentials));
 
 	/* Get the return_url parameter from the not-registered url */
 	uri_parts = g_strsplit (nrrd->not_registered_url, "?", 0);
@@ -336,10 +354,10 @@ got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar
 				    g_str_has_prefix (uri_args[i], "next")) {
 					this_arg = g_strsplit (uri_args[i], "=", 2);
 
-					url_to_use = g_strdup_printf ("%s%s?returnUrl=%s&forward_on_to_url=%s",
+					url_to_use = g_strdup_printf ("%s%s?forward_on_to_url=%s",
 								      nrrd->music_store->priv->base_url,
 								      U1_STORE_URL,
-								      this_arg[1], this_arg[1]);
+								      this_arg[1]);
 					g_strfreev (this_arg);
 					break;
 				}
@@ -354,14 +372,17 @@ got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar
 
 	/* Now sign the URL and load the new page in the view */
 	real_url = oauth_sign_url2 (url_to_use, NULL, OA_HMAC, "GET",
-				    "ubuntuone", "hammertime",
+				    oauth_consumer_token, oauth_consumer_secret,
 				    oauth_token, oauth_token_secret);
 	webkit_web_view_open (WEBKIT_WEB_VIEW (nrrd->music_store->priv->web_viewer), real_url);
 
 	/* Free memory */
-	g_object_unref (G_OBJECT (proxy));
+	g_signal_handlers_disconnect_by_func (auth, got_new_credentials_cb, nrrd);
+
 	g_free (url_to_use);
 	g_free (real_url);
+	g_free (oauth_consumer_token);
+	g_free (oauth_consumer_secret);
 	g_free (oauth_token);
 	g_free (oauth_token_secret);
 
@@ -369,26 +390,36 @@ got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar
 	g_free (nrrd);
 }
 
-static void
-authorization_denied_cb (DBusGProxy *proxy, gpointer user_data)
+static void authorization_cancelled_cb (SyncdaemonAuthentication *auth, gpointer user_data)
 {
 	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;
 
+	g_debug ("SSO authentication cancelled");
+
 	/* Free memory */
+	g_signal_handlers_disconnect_by_func (auth, authorization_cancelled_cb, nrrd);
 	g_free (nrrd->not_registered_url);
 	g_free (nrrd);
+
 }
 
 static void
-oauth_error_cb (DBusGProxy *proxy, const gchar *message, gpointer user_data)
+credentials_error_cb (SyncdaemonAuthentication *auth, const gchar *error, gpointer user_data)
 {
+	GtkWidget *dialog;
 	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;
 
-	g_debug ("OAuth error: %s", message);
+	g_debug ("SSO authentication error: %s", error);
 
-	/* Free memory */
-	g_free (nrrd->not_registered_url);
-	g_free (nrrd);
+	/* Show an error message */
+	dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (nrrd->music_store))),
+					 0,
+					 GTK_MESSAGE_ERROR,
+					 GTK_BUTTONS_CLOSE,
+					 _("Error while getting credentials:\n%s"),
+					 error);
+	gtk_dialog_run (GTK_DIALOG (dialog));
+	gtk_widget_destroy (dialog);
 }
 
 typedef struct {
@@ -429,6 +460,30 @@ fire_authentication (U1MusicStore *music_store, WebKitNetworkRequest *request, W
 {
 	NotRegisteredReplacementData *nrrd;
 	gchar *internal_uri;
+	SyncdaemonAuthentication *auth;
+
+	if (music_store->priv->auth_retries > MAXIMUM_AUTH_RETRIES) {
+		GtkWidget *dialog;
+		gchar *home_url;
+
+		/* Tell the user authentication has failed */
+		dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (music_store))),
+						 0,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_CLOSE,
+						 _("Authentication to the music store failed"));
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+
+		/* And move back to the store's home page */
+		home_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_NOT_LOGGED_IN_STORE_URL);
+		webkit_web_view_open (WEBKIT_WEB_VIEW (music_store->priv->web_viewer), home_url);
+		g_free (home_url);
+
+		return;
+	}
+
+	music_store->priv->auth_retries++;
 
 	g_debug ("firing authentication");
 
@@ -440,12 +495,15 @@ fire_authentication (U1MusicStore *music_store, WebKitNetworkRequest *request, W
 	nrrd->music_store = music_store;
 	nrrd->not_registered_url = g_strdup (webkit_network_request_get_uri (request));
 
-	authentication_set_callbacks (G_CALLBACK (authorization_denied_cb),
-				      G_CALLBACK (authorization_denied_cb),
-				      G_CALLBACK (got_new_dbus_credentials_cb),
-				      G_CALLBACK (oauth_error_cb),
-				      nrrd);
-	authentication_login ();
+	auth = syncdaemon_daemon_get_authentication (music_store->priv->syncdaemon);
+
+	g_signal_connect (auth, "credentials_found",
+			  G_CALLBACK (got_new_credentials_cb), nrrd);
+	g_signal_connect (auth, "authorization_cancelled",
+			  G_CALLBACK (authorization_cancelled_cb), nrrd);
+	g_signal_connect (auth, "error",
+			  G_CALLBACK (credentials_error_cb), nrrd);
+	syncdaemon_authentication_login_or_register (auth);
 }
 
 static gboolean
@@ -484,7 +542,8 @@ navigation_requested_cb (WebKitWebView *web_view,
 		   g_strrstr (parsed_uri->query, "next=") != NULL) {
 		fire_authentication (music_store, request, policy_decision);
 	} else if (g_str_has_prefix (parsed_uri->path, U1_TOKEN_INVALID_URL)) {
-		authentication_clear_tokens ();
+		syncdaemon_authentication_clear_token (syncdaemon_daemon_get_authentication (music_store->priv->syncdaemon));
+		music_store->priv->auth_retries = 0;
 		fire_authentication (music_store, request, policy_decision);
 	} else if (g_strrstr (parsed_uri->path, "library.aspx") != NULL ||
 		 g_str_has_prefix (parsed_uri->path, "/music/store/library")) {
@@ -637,7 +696,7 @@ parse_html_node (U1MusicStore *music_store, WebKitWebView *web_view, htmlNodePtr
 			gchar *script = NULL, *full_path;
 			GString *escaped_prop;
 			gint i;
-			SyncdaemonDownload *download;
+			SyncdaemonTransferInfo *download;
 
 			prop = xmlGetProp (subnode, "data-path");
 			if (prop == NULL)
@@ -659,15 +718,19 @@ parse_html_node (U1MusicStore *music_store, WebKitWebView *web_view, htmlNodePtr
 			download = g_hash_table_lookup (current_downloads, full_path);
 			if (download != NULL) {
 				gdouble percent;
+				glong total_size, bytes_read;
 
-				if (download->deflated_size != 0)
-					percent = (gdouble) download->n_bytes_read / (gdouble) download->deflated_size;
+				total_size = syncdaemon_transfer_info_get_total_size (download);
+				bytes_read = syncdaemon_transfer_info_get_bytes_transferred (download);
+
+				if (total_size != 0)
+					percent = (gdouble) bytes_read / (gdouble) total_size;
 				else
 					percent = 0.0;
 				script = g_strdup_printf ("setProgressBar('%s', %.2f, '%.1fMB of %.1fMB downloaded');",
 							  escaped_prop->str, percent,
-							  (gdouble) download->n_bytes_read / (gdouble) 1024 / (gdouble) 1024,
-							  (gdouble) download->deflated_size / (gdouble) 1024 / (gdouble) 1024);
+							  (gdouble) bytes_read / (gdouble) 1024 / (gdouble) 1024,
+							  (gdouble) total_size / (gdouble) 1024 / (gdouble) 1024);
 			} else {
 				/* If the file exists, the download is done */
 				if (g_file_test (full_path, G_FILE_TEST_EXISTS)) {
@@ -713,24 +776,47 @@ update_status_in_library_page (U1MusicStore *music_store, WebKitWebView *web_vie
 	WebKitWebDataSource *data_source;
 	htmlDocPtr html_doc;
 	htmlNodePtr root_node;
+	GSList *list;
+	SyncdaemonInterface *interface;
 	GHashTable *current_downloads;
 
+	/* Only continue if we can make calls to SyncDaemon */
+	if (!syncdaemon_daemon_is_ready (music_store->priv->syncdaemon))
+		return;
+
 	/* Retrieve contents of the page */
 	data_source = webkit_web_frame_get_data_source (frame);
 	html = webkit_web_data_source_get_data (data_source);
 
 	/* Get list of current downloads */
-	current_downloads = syncdaemon_get_current_downloads ();
-	if (current_downloads != NULL) {
-		/* Parse HTML */
-		html_doc = htmlParseDoc (html->str, NULL);
-		root_node = xmlDocGetRootElement (html_doc);
-		parse_html_node (music_store, web_view, root_node, current_downloads);
-
-		/* Free memory */
-		xmlFreeDoc (html_doc);
-		g_hash_table_destroy (current_downloads);
+	interface = syncdaemon_daemon_get_status_interface (music_store->priv->syncdaemon);
+	if (interface == NULL) {
+		g_warning ("Could not get /status interface for SyncDaemon");
+		return;
 	}
+
+        list = syncdaemon_status_interface_get_current_downloads (SYNCDAEMON_STATUS_INTERFACE (interface));
+
+	current_downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, (GDestroyNotify) g_object_unref);
+
+	while (list != NULL) {
+		SyncdaemonTransferInfo *transfer_info = (SyncdaemonTransferInfo *) list->data;
+
+		g_hash_table_insert (current_downloads,
+				     g_strdup (syncdaemon_transfer_info_get_path (transfer_info)),
+				     g_object_ref (G_OBJECT (transfer_info)));
+
+		list = g_slist_remove (list, transfer_info);
+	}
+
+	/* Parse HTML */
+	html_doc = htmlParseDoc (html->str, NULL);
+	root_node = xmlDocGetRootElement (html_doc);
+	parse_html_node (music_store, web_view, root_node, current_downloads);
+
+	/* Free memory */
+	xmlFreeDoc (html_doc);
+	g_hash_table_destroy (current_downloads);
 }
 
 static gboolean
@@ -781,7 +867,7 @@ load_finished_cb (WebKitWebView *web_view, WebKitWebFrame *frame, gpointer user_
 		update_status_in_library_page (music_store, web_view, frame);
 
 		/* Poll syncdaemon for downloads while the page is loaded */
-		music_store->priv->watch_id = g_timeout_add (5000, (GSourceFunc) poll_downloads_cb, music_store);
+		music_store->priv->watch_id = g_timeout_add (1000, (GSourceFunc) poll_downloads_cb, music_store);
 	}
 
 	g_free (tmp_url);
@@ -880,6 +966,7 @@ u1_music_store_init (U1MusicStore *music_store)
 
 	music_store->priv = g_new0 (U1MusicStorePrivate, 1);
 	music_store->priv->watched_downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+	music_store->priv->syncdaemon = syncdaemon_daemon_new ();
 
 	/* Initialize DBus */
 	music_store->priv->bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
@@ -901,10 +988,6 @@ u1_music_store_init (U1MusicStore *music_store)
 					   G_TYPE_STRING,
 					   G_TYPE_INVALID);
 
-	/* Initialize services */
-	syncdaemon_init ();
-	authentication_init ();
-
 	/* Create web viewer object */
 	music_store->priv->scroll = gtk_scrolled_window_new (NULL, NULL);
 	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (music_store->priv->scroll),
@@ -998,3 +1081,46 @@ u1_music_store_get_library_location (U1MusicStore *music_store)
 
 	return (const gchar *) full_path;
 }
+
+/**
+ * u1_music_store_load_store_link:
+ * @music_store: A #U1MusicStore object
+ * @url: The URL to load
+ *
+ * Load a music store link on the given #U1MusicStore object, passing through
+ * authentication of the user.
+ */
+void
+u1_music_store_load_store_link (U1MusicStore *music_store, const gchar *url)
+{
+	gchar *real_url, *oauth_consumer_token, *oauth_consumer_secret, *oauth_token, *oauth_token_secret;
+
+	g_return_if_fail (U1_IS_MUSIC_STORE (music_store));
+	g_return_if_fail (url != NULL);
+
+	real_url = g_strdup_printf ("%s%s?forward_on_to_url=%s",
+				    music_store->priv->base_url,
+				    U1_STORE_URL,
+				    url);
+
+	/* Sign the URL if we have OAuth tokens */
+	get_credentials (music_store,
+			 &oauth_consumer_token,
+			 &oauth_consumer_secret,
+			 &oauth_token,
+			 &oauth_token_secret);
+	if (oauth_consumer_token != NULL && oauth_consumer_secret != NULL &&
+	    oauth_token != NULL && oauth_token_secret != NULL) {
+		gchar *signed_url;
+
+		signed_url = oauth_sign_url2 (real_url, NULL, OA_HMAC, "GET",
+					      oauth_consumer_token, oauth_consumer_secret,
+					      oauth_token, oauth_token_secret);
+		webkit_web_view_open (WEBKIT_WEB_VIEW (music_store->priv->web_viewer), signed_url);
+
+		g_free (signed_url);
+	} else
+		webkit_web_view_open (WEBKIT_WEB_VIEW (music_store->priv->web_viewer), real_url);
+
+	g_free (real_url);
+}
